@@ -1,6 +1,6 @@
 // This contains the implementation of the pyqtBoundSignal type.
 //
-// Copyright (c) 2012 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2014 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt.
 // 
@@ -59,7 +59,7 @@ static PyObject *connect(qpycore_pyqtBoundSignal *bs, QObject *qrx,
         const char *slot, Qt::ConnectionType type);
 static PyObject *disconnect(qpycore_pyqtBoundSignal *bs, QObject *qrx,
         const char *slot);
-static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
+static QObject *get_receiver(qpycore_pyqtBoundSignal *bs, PyObject *slot_obj,
         QByteArray &name);
 static QByteArray slot_signature_from_decorations(Chimera::Signature *signal,
         PyObject *decorations, int nr_args);
@@ -71,10 +71,12 @@ static QByteArray slot_signature(Chimera::Signature *signal,
 
 // Doc-strings.
 PyDoc_STRVAR(pyqtBoundSignal_connect_doc,
-"connect(slot[, type=Qt.AutoConnection])\n"
+"connect(slot, type=Qt.AutoConnection, no_receiver_check=False)\n"
 "\n"
 "slot is either a Python callable or another signal.\n"
-"type is a Qt.ConnectionType");
+"type is a Qt.ConnectionType.\n"
+"no_receiver_check is True to disable the check that the receiver's C++\n"
+"instance still exists when the signal is emitted.\n");
 
 PyDoc_STRVAR(pyqtBoundSignal_disconnect_doc,
 "disconnect([slot])\n"
@@ -173,6 +175,9 @@ PyTypeObject qpycore_pyqtBoundSignal_Type = {
 #if PY_VERSION_HEX >= 0x02060000
     0,                      /* tp_version_tag */
 #endif
+#if PY_VERSION_HEX >= 0x03040000
+    0,                      /* tp_finalize */
+#endif
 };
 
 
@@ -181,7 +186,24 @@ static PyObject *pyqtBoundSignal_get_doc(PyObject *self, void *)
 {
     qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)self;
 
-    return qpycore_get_signal_doc(bs->unbound_signal);
+    const char *docstring = bs->unbound_signal->docstring;
+
+    if (!docstring)
+    {
+        Py_INCREF(Py_None);
+        return Py_None;
+    }
+
+    if (*docstring == '\1')
+        ++docstring;
+
+    return
+#if PY_MAJOR_VERSION >= 3
+        PyUnicode_FromString
+#else
+        PyString_FromString
+#endif
+            (docstring);
 }
 
 
@@ -196,7 +218,7 @@ static PyObject *pyqtBoundSignal_get_signal(PyObject *self, void *)
 #else
         PyString_FromString
 #endif
-            (bs->bound_overload->signature.constData());
+            (bs->unbound_signal->signature->signature.constData());
 }
 
 
@@ -205,7 +227,7 @@ static PyObject *pyqtBoundSignal_repr(PyObject *self)
 {
     qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)self;
 
-    QByteArray name = Chimera::Signature::name(bs->bound_overload->signature);
+    QByteArray name = bs->unbound_signal->signature->name();
 
     return
 #if PY_MAJOR_VERSION >= 3
@@ -234,26 +256,25 @@ static void pyqtBoundSignal_dealloc(PyObject *self)
 {
     qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)self;
 
-    Py_XDECREF(bs->unbound_signal);
+    Py_XDECREF((PyObject *)bs->unbound_signal);
 
     Py_TYPE(self)->tp_free(self);
 }
 
 
 // Create a bound signal.
-PyObject *qpycore_pyqtBoundSignal_New(PyObject *unbound_signal,
-        PyObject *bound_pyobject, QObject *bound_qobject, int signal_index)
+PyObject *qpycore_pyqtBoundSignal_New(qpycore_pyqtSignal *unbound_signal,
+        PyObject *bound_pyobject, QObject *bound_qobject)
 {
-    qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)PyType_GenericAlloc(&qpycore_pyqtBoundSignal_Type, 0);
+    qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)PyType_GenericNew(&qpycore_pyqtBoundSignal_Type, 0, 0);
 
     if (bs)
     {
-        Py_INCREF(unbound_signal);
+        Py_INCREF((PyObject *)unbound_signal);
         bs->unbound_signal = unbound_signal;
 
         bs->bound_pyobject = bound_pyobject;
         bs->bound_qobject = bound_qobject;
-        bs->bound_overload = ((qpycore_pyqtSignal *)unbound_signal)->overloads->at(signal_index);
     }
 
     return (PyObject *)bs;
@@ -266,15 +287,15 @@ static PyObject *pyqtBoundSignal_mp_subscript(PyObject *self,
 {
     qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)self;
 
-    int idx = qpycore_signal_index((qpycore_pyqtSignal *)bs->unbound_signal,
-            subscript, "a bound signal type argument");
+    qpycore_pyqtSignal *ps = qpycore_find_signal(bs->unbound_signal, subscript,
+            "a bound signal type argument");
 
-    if (idx < 0)
+    if (!ps)
         return 0;
 
     // Create a new bound signal.
-    return qpycore_pyqtBoundSignal_New(bs->unbound_signal, bs->bound_pyobject,
-            bs->bound_qobject, idx);
+    return qpycore_pyqtBoundSignal_New(ps, bs->bound_pyobject,
+            bs->bound_qobject);
 }
 
 
@@ -287,18 +308,22 @@ static PyObject *pyqtBoundSignal_connect(PyObject *self, PyObject *args,
     static const char *kwds[] = {
         "slot",
         "type",
+        "no_receiver_check",
         0
     };
 
     PyObject *slot_obj, *type_obj = 0;
+    int no_receiver_check = 0;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwd_args,
-#if PY_VERSION_HEX >= 0x02050000
-                "O|O:connect",
+#if PY_VERSION_HEX >= 0x03030000
+                "O|Op:connect",
+#elif PY_VERSION_HEX >= 0x02050000
+                "O|Oi:connect",
 #else
-                const_cast<char *>("O|O:connect"),
+                const_cast<char *>("O|Oi:connect"),
 #endif
-                const_cast<char **>(kwds), &slot_obj, &type_obj))
+                const_cast<char **>(kwds), &slot_obj, &type_obj, &no_receiver_check))
         return 0;
 
     Qt::ConnectionType type = Qt::AutoConnection;
@@ -332,7 +357,8 @@ static PyObject *pyqtBoundSignal_connect(PyObject *self, PyObject *args,
         }
 
         return connect(bs, slot_bs->bound_qobject,
-                slot_bs->bound_overload->signature.constData(), type);
+                slot_bs->unbound_signal->signature->signature.constData(),
+                type);
     }
 
     // Make sure the slot is callable.
@@ -348,7 +374,7 @@ static PyObject *pyqtBoundSignal_connect(PyObject *self, PyObject *args,
     // See if the slot can be used directly (ie. it wraps a Qt slot) or if it
     // needs a proxy.
     QByteArray rx_name;
-    QObject *rx_qobj = get_receiver(bs->bound_overload, slot_obj, rx_name);
+    QObject *rx_qobj = get_receiver(bs, slot_obj, rx_name);
 
     if (PyErr_Occurred())
         return 0;
@@ -360,11 +386,32 @@ static PyObject *pyqtBoundSignal_connect(PyObject *self, PyObject *args,
     PyQtProxy *proxy;
     const char *member;
 
+#if QT_VERSION >= 0x040600
+    if ((type & Qt::UniqueConnection) == Qt::UniqueConnection)
+    {
+        proxy = PyQtProxy::findSlotProxy(bs->bound_qobject,
+                bs->unbound_signal->signature->signature.constData(), slot_obj,
+                0, &member);
+
+        if (proxy)
+        {
+            // We give more information than we could if it was a Qt slot but
+            // to be consistent we raise a TypeError even though it's not the
+            // most appropriate for the type of error.
+            PyErr_SetString(PyExc_TypeError, "connection is not unique");
+            return 0;
+        }
+    }
+#endif
+
     Py_BEGIN_ALLOW_THREADS
 
     proxy = new PyQtProxy(bs, slot_obj, &member);
 
-    if (proxy->real_slot.signature)
+    if (no_receiver_check)
+        proxy->disableReceiverCheck();
+
+    if (proxy->metaObject())
     {
         if (rx_qobj)
             proxy->moveToThread(rx_qobj->thread());
@@ -389,10 +436,11 @@ static PyObject *connect(qpycore_pyqtBoundSignal *bs, QObject *qrx,
         const char *slot, Qt::ConnectionType type)
 {
     bool ok;
+    Chimera::Signature *signature = bs->unbound_signal->signature;
 
     Py_BEGIN_ALLOW_THREADS
-    ok = QObject::connect(bs->bound_qobject,
-            bs->bound_overload->signature.constData(), qrx, slot, type);
+    ok = QObject::connect(bs->bound_qobject, signature->signature.constData(),
+            qrx, slot, type);
     Py_END_ALLOW_THREADS
 
     if (!ok)
@@ -400,8 +448,7 @@ static PyObject *connect(qpycore_pyqtBoundSignal *bs, QObject *qrx,
         QByteArray rx_name = Chimera::Signature::name(slot);
 
         PyErr_Format(PyExc_TypeError, "connect() failed between %s and %s()",
-                bs->bound_overload->py_signature.constData(),
-                rx_name.constData() + 1);
+                signature->py_signature.constData(), rx_name.constData() + 1);
 
         return 0;
     }
@@ -417,6 +464,7 @@ static PyObject *pyqtBoundSignal_disconnect(PyObject *self, PyObject *args)
     qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)self;
 
     PyObject *slot_obj = 0, *res_obj;
+    Chimera::Signature *signature = bs->unbound_signal->signature;
 
 #if PY_VERSION_HEX >= 0x02050000
     if (!PyArg_ParseTuple(args, "|O:disconnect", &slot_obj))
@@ -431,7 +479,7 @@ static PyObject *pyqtBoundSignal_disconnect(PyObject *self, PyObject *args)
         res_obj = disconnect(bs, 0, 0);
 
         PyQtProxy::deleteSlotProxies(bs->bound_qobject,
-                bs->bound_overload->signature.constData());
+                signature->signature.constData());
 
         return res_obj;
     }
@@ -442,7 +490,7 @@ static PyObject *pyqtBoundSignal_disconnect(PyObject *self, PyObject *args)
         qpycore_pyqtBoundSignal *slot_bs = (qpycore_pyqtBoundSignal *)slot_obj;
 
         return disconnect(bs, slot_bs->bound_qobject,
-                slot_bs->bound_overload->signature.constData());
+                slot_bs->unbound_signal->signature->signature.constData());
     }
 
     if (!PyCallable_Check(slot_obj))
@@ -457,7 +505,7 @@ static PyObject *pyqtBoundSignal_disconnect(PyObject *self, PyObject *args)
     // See if the slot has been used directly (ie. it wraps a Qt slot) or if it
     // has a proxy.
     QByteArray rx_name;
-    QObject *rx_qobj = get_receiver(bs->bound_overload, slot_obj, rx_name);
+    QObject *rx_qobj = get_receiver(bs, slot_obj, rx_name);
 
     if (PyErr_Occurred())
         return 0;
@@ -467,7 +515,7 @@ static PyObject *pyqtBoundSignal_disconnect(PyObject *self, PyObject *args)
 
     const char *member;
     PyQtProxy *proxy = PyQtProxy::findSlotProxy(bs->bound_qobject,
-            bs->bound_overload->signature.constData(), slot_obj, 0, &member);
+            signature->signature.constData(), slot_obj, 0, &member);
 
     if (!proxy)
     {
@@ -489,16 +537,17 @@ static PyObject *pyqtBoundSignal_disconnect(PyObject *self, PyObject *args)
 static PyObject *disconnect(qpycore_pyqtBoundSignal *bs, QObject *qrx,
         const char *slot)
 {
+    Chimera::Signature *signature = bs->unbound_signal->signature;
     bool ok;
 
     Py_BEGIN_ALLOW_THREADS
     ok = QObject::disconnect(bs->bound_qobject,
-            bs->bound_overload->signature.constData(), qrx, slot);
+            signature->signature.constData(), qrx, slot);
     Py_END_ALLOW_THREADS
 
     if (!ok)
     {
-        QByteArray tx_name = bs->bound_overload->name();
+        QByteArray tx_name = signature->name();
 
         if (slot)
         {
@@ -530,9 +579,10 @@ static PyObject *pyqtBoundSignal_emit(PyObject *self, PyObject *args)
 
     if (!bs->bound_qobject->signalsBlocked())
     {
-        int signal_index = bs->bound_qobject->metaObject()->indexOfSignal(bs->bound_overload->signature.constData() + 1);
+        Chimera::Signature *signature = bs->unbound_signal->signature;
+        int mo_index = bs->bound_qobject->metaObject()->indexOfSignal(signature->signature.constData() + 1);
 
-        if (signal_index < 0)
+        if (mo_index < 0)
         {
             PyErr_Format(PyExc_AttributeError,
                     "signal was not defined in the first super-class of class '%s'",
@@ -540,8 +590,105 @@ static PyObject *pyqtBoundSignal_emit(PyObject *self, PyObject *args)
             return 0;
         }
 
-        if (!qpycore_emit(bs->bound_qobject, signal_index, bs->bound_overload, args))
+        // Use the docstring if there is one and it is auto-generated.
+        const char *docstring = bs->unbound_signal->docstring;
+
+        if (!docstring || *docstring != '\1')
+        {
+            docstring = signature->py_signature.constData();
+        }
+        else
+        {
+            // Skip the auto-generated marker.
+            ++docstring;
+        }
+
+#if QT_VERSION < 0x050000
+        if (!qpycore_emit(bs->bound_qobject, mo_index, signature, docstring, args))
             return 0;
+#else
+        // Qt5 changes the way signals with optional arguments are handled in
+        // that the version with all arguments used is always used when making
+        // connections.  (Note that it's possible Qt4 does this as well but we
+        // have never noticed because Qt4 only has one signal with optional
+        // arguments.)  We can't do what PyQt5 does because it exploits the
+        // fact that Qt5 signals are public rather than protected.  We could do
+        // what PyQt3 does (by creating a public wrapper in the derived class)
+        // but that is a bit complicated.  We could implement the parsing of
+        // optional arguments in Chimera, but that is also a bit complicated.
+        // Instead, at this late stage in the life of PyQt4, we keep the code
+        // changes to a minimum and handle it with a big hack.  Each
+        // problematic signal has a number attached to it which specifies
+        // exactly what hack is needed.
+
+        PyObject *hacked_args;
+        int hack = bs->unbound_signal->hack;
+
+        // If all the arguments has been provided then we don't need the hack.
+        if (PyTuple_GET_SIZE(args) == signature->parsed_arguments.size())
+        {
+            hack = 0;
+        }
+        else
+        {
+            hacked_args = PyTuple_New(PyTuple_GET_SIZE(args) + 1);
+
+            if (!hacked_args)
+                return 0;
+
+            for (SIP_SSIZE_T i = 0; i < PyTuple_GET_SIZE(args); ++i)
+            {
+                PyObject *arg = PyTuple_GET_ITEM(args, i);
+
+                Py_INCREF(arg);
+                PyTuple_SET_ITEM(hacked_args, i, arg);
+            }
+        }
+
+        switch (hack)
+        {
+        case 1:
+            // Supply the missing None.
+            Py_INCREF(Py_None);
+            PyTuple_SET_ITEM(hacked_args, PyTuple_GET_SIZE(args), Py_None);
+            break;
+
+        case 2:
+            // Supply the missing [].
+            {
+                PyObject *empty_list = PyList_New(0);
+
+                if (!empty_list)
+                {
+                    Py_DECREF(hacked_args);
+                    return 0;
+                }
+
+                PyTuple_SET_ITEM(hacked_args, PyTuple_GET_SIZE(args),
+                        empty_list);
+            }
+
+            break;
+
+        case 3:
+            // Supply the missing False.
+            Py_INCREF(Py_False);
+            PyTuple_SET_ITEM(hacked_args, PyTuple_GET_SIZE(args), Py_False);
+            break;
+
+        default:
+            hacked_args = args;
+            Py_INCREF(hacked_args);
+        }
+
+        bool ok = qpycore_emit(bs->bound_qobject, mo_index, signature,
+                docstring, hacked_args);
+
+        Py_DECREF(hacked_args);
+
+        if (!ok)
+            return 0;
+#endif
     }
 
     Py_INCREF(Py_None);
@@ -552,12 +699,13 @@ static PyObject *pyqtBoundSignal_emit(PyObject *self, PyObject *args)
 // Get the receiver QObject from the slot (if there is one) and its signature
 // (if it wraps a Qt slot).  A Python exception will be raised if there was an
 // error.
-static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
+static QObject *get_receiver(qpycore_pyqtBoundSignal *bs, PyObject *slot_obj,
         QByteArray &name)
 {
     PyObject *rx_self, *decorations;
     QByteArray rx_name;
     bool try_qt_slot;
+    Chimera::Signature *signature = bs->unbound_signal->signature;
 
     decorations = 0;
 
@@ -589,6 +737,8 @@ static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
         {
             try_qt_slot = false;
         }
+
+        Py_XINCREF(rx_self);
     }
     else if (PyCFunction_Check(slot_obj))
     {
@@ -602,10 +752,63 @@ static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
             rx_name.chop(1);
 
         try_qt_slot = true;
+
+        Py_XINCREF(rx_self);
     }
     else
     {
-        rx_self = 0;
+        static PyObject *partial = 0;
+
+        // Get the functools.partial type object if we haven't already got it.
+        if (!partial)
+        {
+            PyObject *functools = PyImport_ImportModule("functools");
+
+            if (functools)
+            {
+                partial = PyObject_GetAttrString(functools, "partial");
+                Py_DECREF(functools);
+            }
+        }
+
+        // If we know about functools.partial then remove the outer partials to
+        // get to the original function.
+        if (partial && PyObject_IsInstance(slot_obj, partial))
+        {
+            PyObject *func = slot_obj;
+
+            Py_INCREF(func);
+
+            do
+            {
+                PyObject *subfunc = PyObject_GetAttrString(func, "func");
+
+                Py_DECREF(func);
+
+                // This should never happen.
+                if (!subfunc)
+                    return 0;
+
+                func = subfunc;
+            }
+            while (PyObject_IsInstance(func, partial));
+
+            if (PyMethod_Check(func))
+                rx_self = PyMethod_GET_SELF(func);
+            else if (PyCFunction_Check(func))
+                rx_self = PyCFunction_GET_SELF(func);
+            else
+                rx_self = 0;
+
+            Py_XINCREF(rx_self);
+            Py_DECREF(func);
+
+            try_qt_slot = false;
+        }
+        else
+        {
+            rx_self = 0;
+        }
     }
  
     if (!rx_self)
@@ -615,6 +818,8 @@ static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
     void *rx = sipForceConvertToType(rx_self, sipType_QObject, 0,
             SIP_NO_CONVERTORS, 0, &iserr);
 
+    Py_DECREF(rx_self);
+
     PyErr_Clear();
 
     if (iserr)
@@ -626,7 +831,7 @@ static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
     // of them) then use it.  Otherwise we will fallback to using a proxy.
     if (try_qt_slot)
     {
-        for (int ol = overload->parsed_arguments.count(); ol >= 0; --ol)
+        for (int ol = signature->parsed_arguments.count(); ol >= 0; --ol)
         {
             // If there are decorations then we compare the signal's signature
             // against them so that we distinguish between Python types that
@@ -634,10 +839,10 @@ static QObject *get_receiver(Chimera::Signature *overload, PyObject *slot_obj,
             // distinction.  If there are no decorations then let Qt determine
             // if a slot is available.
             if (decorations)
-                name = slot_signature_from_decorations(overload, decorations,
+                name = slot_signature_from_decorations(signature, decorations,
                         ol);
             else
-                name = slot_signature_from_metaobject(overload,
+                name = slot_signature_from_metaobject(signature,
                         rx_qobj->metaObject(), rx_name, ol);
 
             if (!name.isEmpty())
@@ -771,14 +976,14 @@ const char *qpycore_pyqtsignal_get_parts(PyObject *sig_obj, QObject **qtx)
         if (qtx)
             *qtx = bs->bound_qobject;
 
-        return bs->bound_overload->signature.constData();
+        return bs->unbound_signal->signature->signature.constData();
     }
 
     if (PyObject_TypeCheck(sig_obj, &qpycore_pyqtSignal_Type) && !qtx)
     {
         qpycore_pyqtSignal *ps = (qpycore_pyqtSignal *)sig_obj;
 
-        return ps->overloads->at(ps->master_index)->signature.constData();
+        return ps->signature->signature.constData();
     }
 
     return 0;
